import pytest

import datetime

from acondbs.db.sa import sa
from acondbs.models import (
    GitHubUser,
    GitHubToken,
)

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):

    y = app_empty

    user1 = GitHubUser(login="user1", git_hub_id="04:User1")
    token1 = GitHubToken(
        token="39d86487d76a84087f1da599c872dac4473e5f07", scope="", user=user1
    )

    with y.app_context():
        sa.session.add(user1)
        sa.session.add(token1)
        sa.session.commit()

    # map1 -> beam1
    #   |        |
    #   +--------+---> beam2
    #
    # map2
    # map3

    with y.app_context():
        ops.create_product_relation_type(
            type_={
                "type_id": 1,
                "name": "parent",
            },
            reverse={
                "type_id": 2,
                "name": "child",
            },
        )
        ops.commit()

    with y.app_context():
        ops.create_field(
            field_id=1,
            name="contact",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=2,
            name="produced_by",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=3,
            name="date_produced",
            type_=ops.FieldType.Date,
        )
        ops.commit()

    with y.app_context():
        ops.create_product_type(
            type_id=1,
            name="map",
            field_ids=[1, 2, 3],
        )
        ops.create_product_type(
            type_id=2,
            name="beam",
            field_ids=[1, 2, 3],
        )
        ops.commit()

    with y.app_context():
        ops.create_product(
            type_id=1,
            product_id=1,
            name="map1",
            attributes={3: datetime.date(2020, 2, 1)},
            paths=["site1:/path/to/map1", "site2:/another/way/map1"],
        )
        ops.create_product(
            type_id=1,
            product_id=2,
            name="map2",
            attributes={3: datetime.date(2020, 2, 10)},
            paths=["site1:/path/to/map2"],
        )
        ops.create_product(
            type_id=1,
            product_id=3,
            name="map3",
            attributes={3: datetime.date(2020, 3, 19)},
            paths=["site1:/path/to/map3", "site2:/another/way/map3"],
        )
        ops.create_product(
            type_id=2,
            product_id=4,
            name="beam1",
            attributes={3: datetime.date(2020, 2, 5)},
            paths=["site1:/path/to/beam1", "site2:/another/way/beam1"],
        )
        ops.create_product(
            type_id=2,
            product_id=5,
            name="beam2",
            attributes={3: datetime.date(2020, 3, 4)},
            paths=["site1:/path/to/beam2"],
        )
        ops.commit()

    with y.app_context():
        ops.create_product_relation(
            type_id=1,
            self_product_id=4,
            other_product_id=1,
        )
        ops.create_product_relation(
            type_id=1,
            self_product_id=5,
            other_product_id=1,
        )
        ops.create_product_relation(
            type_id=1,
            self_product_id=5,
            other_product_id=4,
        )
        ops.commit()

    yield y


##__________________________________________________________________||



from ..fragments import FRAGMENT_LOG

MUTATION_CREATE_LOG = '''
mutation CreateLog($input: CreateLogInput!) {
  createLog(input: $input) {
    ok
    log {
      ...fragmentLog
    }
  }
}
''' + FRAGMENT_LOG



"""add fields to github_admin_app_token

Revision ID: 63033c01def0
Revises: f92f99aa6fdd
Create Date: 2020-12-25 13:51:10.399740

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '63033c01def0'
down_revision = 'f92f99aa6fdd'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_admin_app_token', schema=None) as batch_op:
        batch_op.add_column(sa.Column('scope', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('user_id', sa.Integer(), nullable=False))
        batch_op.create_foreign_key(None, 'github_users', ['user_id'], ['user_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_admin_app_token', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_column('user_id')
        batch_op.drop_column('scope')

    # ### end Alembic commands ###



from acondbs import ops

from acondbs.models import Log


def test_create(app):
    with app.app_context():
        count = Log.query.count()
        model = ops.create_log(level="ERROR", message="An exception occurred")
        assert model.message == "An exception occurred"
        ops.commit()
        id_ = model.id_
        assert id_

    with app.app_context():
        assert Log.query.count() == (count + 1)
        model = Log.query.filter_by(id_=id_).one()
        assert model.message == "An exception occurred"


def test_delete(app):
    with app.app_context():
        model = ops.create_log(level="ERROR", message="To be deleted")
        ops.commit()
        id_ = model.id_

    with app.app_context():
        count = Log.query.count()
        ops.delete_log(id_=id_)
        ops.commit()

    with app.app_context():
        assert Log.query.count() == (count - 1)
        model = Log.query.filter_by(id_=id_).one_or_none()
        assert model is None



from ..fragments import FRAGMENT_PRODUCT

MUTATION_UPDATE_PRODUCT = '''
mutation($productId: Int!, $input: UpdateProductInput!) {
  updateProduct(productId: $productId, input: $input) {
    ok
    product {
      ...fragmentProduct
    }
  }
}
''' + FRAGMENT_PRODUCT



from ..fragments import FRAGMENT_PRODUCT

QUERY_PRODUCT = (
    """
query Product($productId: Int, $typeId: Int, $name: String) {
  product(productId: $productId, typeId: $typeId, name: $name) {
    ...fragmentProduct
  }
}
"""
    + FRAGMENT_PRODUCT
)



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..connection import CountedConnection
from ...models import (
    Product as ProductModel,
    ProductType as ProductTypeModel,
    ProductFilePath as ProductFilePathModel,
    ProductRelation as ProductRelationModel,
    ProductRelationType as ProductRelationTypeModel,
    # FieldType as FieldTypeModel,  # enum
    saEnumFieldType,  # SQLAlchemy Enum
    Field as FieldModel,
    TypeFieldAssociation as TypeFieldAssociationModel,
    AttributeUnicodeText as AttributeUnicodeTextModel,
    AttributeBoolean as AttributeBooleanModel,
    AttributeInteger as AttributeIntegerModel,
    AttributeFloat as AttributeFloatModel,
    AttributeDate as AttributeDateModel,
    AttributeDateTime as AttributeDateTimeModel,
    AttributeTime as AttributeTimeModel,
)
from ..filter_ import PFilterableConnectionField


from graphene_sqlalchemy.enums import _convert_sa_to_graphene_enum


##__________________________________________________________________||
# FieldType is an enum. It is manually converted to a graphene enum
# here. An enum will be usually automatically converted. However, the
# automatic conversion causes an error if the automatic conversion
# of the same enum happens multiple times as described in
# https://github.com/graphql-python/graphene-sqlalchemy/issues/211.

# It is possible to use graphene.Enum.from_enum() as
# FieldType = graphene.Enum.from_enum(FieldTypeModel)

# However, instead, a private function _convert_sa_to_graphene_enum()
# is used here because this function desirably converts field names to
# the upper case, e.g., "UnicodeText" to "UNICODE_TEXT"
FieldType = _convert_sa_to_graphene_enum(saEnumFieldType)


##__________________________________________________________________||
class Product(SQLAlchemyObjectType):
    class Meta:
        model = ProductModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class ProductType(SQLAlchemyObjectType):
    """A product type"""

    class Meta:
        model = ProductTypeModel
        interfaces = (graphene.relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class ProductRelation(SQLAlchemyObjectType):
    """A relation from one product to another"""

    class Meta:
        model = ProductRelationModel
        interfaces = (graphene.relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class ProductRelationType(SQLAlchemyObjectType):
    """A type of relations between products"""

    class Meta:
        model = ProductRelationTypeModel
        interfaces = (graphene.relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class ProductFilePath(SQLAlchemyObjectType):
    class Meta:
        model = ProductFilePathModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class Field(SQLAlchemyObjectType):
    class Meta:
        model = FieldModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory
        exclude_fields = ("type_",)

    # Set "type_" manually to avoid an issue on enum as suggested in
    # https://github.com/graphql-python/graphene-sqlalchemy/pull/98#issuecomment-481497115
    # see also https://github.com/graphql-python/graphene-sqlalchemy/issues/211

    type_ = FieldType()


class TypeFieldAssociation(SQLAlchemyObjectType):
    class Meta:
        model = TypeFieldAssociationModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeUnicodeText(SQLAlchemyObjectType):
    """A unicode text attribute of a product"""

    class Meta:
        model = AttributeUnicodeTextModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeBoolean(SQLAlchemyObjectType):
    """A boolean attribute of a product"""

    class Meta:
        model = AttributeBooleanModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeInteger(SQLAlchemyObjectType):
    """An integer attribute of a product"""

    class Meta:
        model = AttributeIntegerModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeFloat(SQLAlchemyObjectType):
    """A float attribute of a product"""

    class Meta:
        model = AttributeFloatModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeDate(SQLAlchemyObjectType):
    """A date attribute of a product"""

    class Meta:
        model = AttributeDateModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeDateTime(SQLAlchemyObjectType):
    """A date time attribute of a product"""

    class Meta:
        model = AttributeDateTimeModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


class AttributeTime(SQLAlchemyObjectType):
    """A time attribute of a product"""

    class Meta:
        model = AttributeTimeModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_graphql 1'] = {
    'data': {
        'webConfig': {
            'id_': '1',
            'json': '{"headTitle": "Product DB (test)", "toolbarTitle": "Product DB (test)", "devtoolLoadingstate": true, "productCreationDialog": false, "productUpdateDialog": true, "productDeletionDialog": false}'
        }
    }
}



QUERY_ALL_PRODUCTS_TOTAL_COUNT = """
query AllProductsTotalCount(
  $filters: ProductFilter
  $sort: [ProductSortEnum]
  $first: Int
) {
  allProducts(filters: $filters, sort: $sort, first: $first) {
    totalCount
  }
}
"""



import pytest

from acondbs import ops
from acondbs.models import FieldType, Field


##__________________________________________________________________||
params = [
    pytest.param(FieldType.Float, id="by-enum"),
    pytest.param(4, id="by-int"),
]


@pytest.mark.parametrize("type_", params)
def test_create(app_empty, type_):
    app = app_empty

    with app.app_context():
        model = ops.create_field(name="new_field", type_=type_)
        assert model.name == "new_field"
        assert model.type_ == FieldType.Float
        ops.commit()
        field_id = model.field_id
        assert field_id

    with app.app_context():
        model = Field.query.filter_by(field_id=field_id).one()
        assert model.name == "new_field"
        assert model.type_ == FieldType.Float
        assert model.field_id == field_id


##__________________________________________________________________||
def test_fixture(app):
    with app.app_context():
        assert Field.query.count() == 9


##__________________________________________________________________||
def test_update(app):

    with app.app_context():
        model = Field.query.filter_by(name="number").one()
        field_id = model.field_id

    with app.app_context():
        model = ops.update_field(field_id=field_id, name="number_renamed")
        ops.commit()

    with app.app_context():
        model = Field.query.filter_by(field_id=field_id).one()
        assert model.name == "number_renamed"
        assert model.field_id == field_id


##__________________________________________________________________||
def test_delete(app):

    with app.app_context():
        model = ops.create_field(name="to_be_deleted", type_=FieldType.Float)
        ops.commit()
        field_id = model.field_id

    with app.app_context():
        count = Field.query.count()
        ret = ops.delete_field(field_id=field_id)
        ops.commit()
        assert ret is None

    with app.app_context():
        model = Field.query.filter_by(field_id=field_id).one_or_none()
        assert model is None
        assert Field.query.count() == count - 1


##__________________________________________________________________||



import datetime
from sqlalchemy.orm import aliased

import sqlparse

import pytest

from acondbs.db.sa import sa

from acondbs.models import (
    Product,
    ProductType,
    FieldType,
    Field,
    TypeFieldAssociation,
    AttributeUnicodeText,
    AttributeBoolean,
    AttributeInteger,
    AttributeFloat,
    AttributeDate,
    AttributeDateTime,
    AttributeTime,
)


##__________________________________________________________________||
params = [
    pytest.param(
        FieldType.UnicodeText,
        AttributeUnicodeText,
        "value1 値１",
        id="UnicodeText",
    ),
    pytest.param(FieldType.Boolean, AttributeBoolean, True, id="Boolean"),
    pytest.param(FieldType.Integer, AttributeInteger, -512442, id="Integer"),
    pytest.param(FieldType.Float, AttributeFloat, 3.14592613, id="Float"),
    pytest.param(
        FieldType.Date, AttributeDate, datetime.date(2020, 2, 1), id="Date"
    ),
    pytest.param(
        FieldType.DateTime,
        AttributeDateTime,
        datetime.datetime(2020, 2, 1, 9, 10, 25),
        id="DateTime",
    ),
    pytest.param(
        FieldType.Time, AttributeTime, datetime.time(9, 10, 25), id="Time"
    ),
]


@pytest.mark.parametrize("field_type, AttributeClass, value", params)
def test_repr(app_empty, field_type, AttributeClass, value):
    app = app_empty  # noqa: F841

    attr1 = AttributeClass(value=value)
    repr(attr1)

    field1 = Field(name="attr1", type_=field_type)
    assoc1 = TypeFieldAssociation(field=field1)
    attr1.type_field_association = assoc1
    attr1.field = field1
    repr(attr1)


@pytest.mark.parametrize("field_type, AttributeClass, value", params)
def test_obj(app_empty, field_type, AttributeClass, value):
    app = app_empty  # noqa: F841
    field = Field(name="field1", type_=field_type)
    assoc = TypeFieldAssociation(field=field)
    attr = AttributeClass(
        type_field_association=assoc, field=field, value=value
    )  # noqa: F841


@pytest.mark.parametrize("field_type, AttributeClass, value", params)
def test_commit(app_empty, field_type, AttributeClass, value):
    app = app_empty
    with app.app_context():
        field = Field(name="field1", type_=field_type)
        assoc = TypeFieldAssociation(field=field)
        product_type = ProductType(name="map")
        product_type.fields = [assoc]
        sa.session.add(product_type)
        sa.session.commit()

    with app.app_context():
        product_type = ProductType.query.filter_by(name="map").one()
        assoc = product_type.fields[0]
        field = assoc.field
        product = Product(name="product1", type_=product_type)
        attr = AttributeClass(
            product=product,
            type_field_association=assoc,
            field=field,
            value=value,
        )
        sa.session.add(attr)
        sa.session.commit()

    with app.app_context():
        attr = AttributeClass.query.one()
        assert attr.__class__ is AttributeClass
        assert attr.type_field_association.field.name == "field1"
        assert attr.type_field_association.field.type_ is field_type
        assert attr.field.name == "field1"
        assert attr.field.type_ is field_type
        assert attr.value == value
        assert getattr(attr.product, AttributeClass.backref_column) == [attr]
        assert getattr(attr.field, AttributeClass.backref_column) == [attr]


##__________________________________________________________________||
def test_filter(app):

    with app.app_context():
        expected = Product.query.filter_by(product_id=1).one()
        print(expected.attributes_unicode_text[0].type_field_association)
        query = (
            Product.query.join(AttributeUnicodeText)
            .join(TypeFieldAssociation)
            .join(Field)
            .filter(
                Field.name == "attr1", AttributeUnicodeText.value == "value1"
            )
        )
        # print(sqlparse.format(str(query), reindent=True))
        actual = query.one()
        assert actual is expected

        # another way
        query = (
            Product.query.join(AttributeUnicodeText)
            .filter_by(value="value1")
            .join(TypeFieldAssociation)
            .join(Field)
            .filter_by(name="attr1")
        )
        # print(sqlparse.format(str(query), reindent=True))
        actual = query.one()
        assert actual is expected


def test_order(app):

    with app.app_context():
        map1 = Product.query.filter_by(product_id=1).one()
        map2 = Product.query.filter_by(product_id=2).one()
        map3 = Product.query.filter_by(product_id=3).one()

        expected = [map3, map2, map1]

        query = (
            Product.query.join(ProductType)
            .filter_by(name="map")
            .join(AttributeDate)
            .join(TypeFieldAssociation)
            .join(Field)
            .filter_by(name="date_produced")
            .order_by(AttributeDate.value.desc())
        )

        # print(sqlparse.format(str(query), reindent=True))

        actual = query.all()

        assert actual == expected


##__________________________________________________________________||
def test_order_nested(app_empty):
    app = app_empty

    field_attr1 = Field(name="attr1", type_=FieldType.UnicodeText)
    field_attr2 = Field(name="attr2", type_=FieldType.UnicodeText)
    Map = ProductType(type_id=1, name="map")
    assoc1 = TypeFieldAssociation(field=field_attr1)
    assoc2 = TypeFieldAssociation(field=field_attr2)
    Map.fields = [assoc1, assoc2]

    map1 = Product(product_id=1, name="map1", type_=Map)
    map2 = Product(product_id=2, name="map2", type_=Map)
    map3 = Product(product_id=3, name="map3", type_=Map)
    map4 = Product(product_id=4, name="map4", type_=Map)
    map5 = Product(product_id=5, name="map5", type_=Map)

    AttributeUnicodeText(
        product=map1,
        type_field_association=assoc1,
        field=field_attr1,
        value="1",
    )
    AttributeUnicodeText(
        product=map2,
        type_field_association=assoc1,
        field=field_attr1,
        value="2",
    )
    AttributeUnicodeText(
        product=map3,
        type_field_association=assoc1,
        field=field_attr1,
        value="1",
    )
    AttributeUnicodeText(
        product=map4,
        type_field_association=assoc1,
        field=field_attr1,
        value="2",
    )
    AttributeUnicodeText(
        product=map5,
        type_field_association=assoc1,
        field=field_attr1,
        value="1",
    )

    AttributeUnicodeText(
        product=map1,
        type_field_association=assoc2,
        field=field_attr2,
        value="b",
    )
    AttributeUnicodeText(
        product=map2,
        type_field_association=assoc2,
        field=field_attr2,
        value="a",
    )
    AttributeUnicodeText(
        product=map3,
        type_field_association=assoc2,
        field=field_attr2,
        value="c",
    )
    AttributeUnicodeText(
        product=map4,
        type_field_association=assoc2,
        field=field_attr2,
        value="b",
    )
    AttributeUnicodeText(
        product=map5,
        type_field_association=assoc2,
        field=field_attr2,
        value="a",
    )

    with app.app_context():
        sa.session.add(Map)
        sa.session.commit()

    with app.app_context():

        map1 = Product.query.filter_by(product_id=1).one()
        map2 = Product.query.filter_by(product_id=2).one()
        map3 = Product.query.filter_by(product_id=3).one()
        map4 = Product.query.filter_by(product_id=4).one()
        map5 = Product.query.filter_by(product_id=5).one()

        expected = [map2, map4, map5, map1, map3]

        # ORM Alias: https://docs.sqlalchemy.org/en/14/tutorial/data_select.html#orm-entity-aliases
        # refer to the same table multiple times
        AliasedAttributeUnicodeText1 = aliased(AttributeUnicodeText)
        AliasedAttributeUnicodeText2 = aliased(AttributeUnicodeText)

        AliasedTypeFieldAssociation1 = aliased(TypeFieldAssociation)
        AliasedTypeFieldAssociation2 = aliased(TypeFieldAssociation)

        AliasedField1 = aliased(Field)
        AliasedField2 = aliased(Field)

        # sort descending order of attr1 then ascending order of attr2
        query = (
            Product.query.join(ProductType)
            .filter_by(name="map")
            .join(AliasedAttributeUnicodeText1)
            .join(AliasedTypeFieldAssociation1)
            .join(AliasedField1)
            .filter_by(name="attr1")
            .order_by(AliasedAttributeUnicodeText1.value.desc())
            .join(
                AliasedAttributeUnicodeText2, Product.attributes_unicode_text
            )
            .join(AliasedTypeFieldAssociation2)
            .join(AliasedField2)
            .filter_by(name="attr2")
            .order_by(AliasedAttributeUnicodeText2.value)
        )

        # print(sqlparse.format(str(query), reindent=True))

        actual = query.all()

        assert actual == expected


##__________________________________________________________________||
def test_delte_orphan_product(app_empty):
    app = app_empty

    with app.app_context():
        field1 = Field(name="field1", type_=FieldType.UnicodeText)
        assoc1 = TypeFieldAssociation(field=field1)
        type1 = ProductType(type_id=1, name="type1", fields=[assoc1])

        product1 = Product(name="product1", type_=type1)
        field1_attribute_class = field1.type_.attribute_class
        attr1 = field1_attribute_class(
            product=product1,
            type_field_association=assoc1,
            field=field1,
            value="value1",
        )

        sa.session.add(product1)
        sa.session.commit()

        product1_id = product1.product_id
        attr1_id = attr1.iid

    with app.app_context():
        product1 = Product.query.filter_by(product_id=product1_id).one()
        sa.session.delete(product1)
        sa.session.commit()

    with app.app_context():
        product1 = Product.query.filter_by(
            product_id=product1_id
        ).one_or_none()
        attr1 = field1_attribute_class.query.filter_by(
            iid=attr1_id
        ).one_or_none()
        assert product1 is None
        assert attr1 is None


def test_delte_orphan_type_field_association_and_field(app_empty):
    app = app_empty

    with app.app_context():
        field1 = Field(name="field1", type_=FieldType.UnicodeText)
        assoc1 = TypeFieldAssociation(field=field1)
        type1 = ProductType(type_id=1, name="type1", fields=[assoc1])

        product1 = Product(name="product1", type_=type1)
        field1_attribute_class = field1.type_.attribute_class
        attr1 = field1_attribute_class(
            product=product1,
            type_field_association=assoc1,
            field=field1,
            value="value1",
        )

        sa.session.add(product1)
        sa.session.commit()

        field1_id = field1.field_id
        assoc1_id = assoc1.iid
        product1_id = product1.product_id
        attr1_id = attr1.iid

    with app.app_context():
        assoc1 = TypeFieldAssociation.query.filter_by(iid=assoc1_id).one()
        field1 = Field.query.filter_by(field_id=field1_id).one()
        sa.session.delete(assoc1)
        sa.session.delete(field1)
        sa.session.commit()

    with app.app_context():
        product1 = Product.query.filter_by(product_id=product1_id).one()
        attr1 = field1_attribute_class.query.filter_by(
            iid=attr1_id
        ).one_or_none()
        assert attr1 is None


##__________________________________________________________________||



"""add column on type_field_association

Revision ID: f97ed2cdfb29
Revises: 8e3f13ce98cb
Create Date: 2021-11-05 11:05:37.683957

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f97ed2cdfb29'
down_revision = '8e3f13ce98cb'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('type_field_association', schema=None) as batch_op:
        batch_op.add_column(sa.Column('order', sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('type_field_association', schema=None) as batch_op:
        batch_op.drop_column('order')

    # ### end Alembic commands ###




import pytest
import unittest.mock as mock

from acondbs.github.call import call_graphql_api

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_requests(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.github.call.requests", y)
    yield y


##__________________________________________________________________||
def test_simple(mock_requests, snapshot):

    query = '{ viewer { login name avatarUrl } }'
    token = 'token-xxx'

    response = {
        "data": {
            "viewer": {
                "login": "octocat",
                "name": "The Octocat",
                "avatarUrl": "https://avatars3.githubusercontent.com/u/583231?u=a59fef2a493e2b67dd13754231daf220c82ba84d&v=4"
            }
        }
    }
    mock_requests.post().json.return_value = response

    r = call_graphql_api(query=query, token=token)

    assert response['data'] == r

    snapshot.assert_match(mock_requests.post.call_args_list)

##__________________________________________________________________||
def test_variables(mock_requests, snapshot):

    query = '''
      query Organization($org_login: String!) {
        organization(login: $org_login) {
          login
          avatarUrl
        }
      }
    '''

    variables = {"org_login": "urban-octo-disco"}

    token = 'token-xxx'

    response = {
        "data": {
            "organization": {
            "login": "urban-octo-disco",
            "avatarUrl": "https://avatars0.githubusercontent.com/u/75631844?v=4"
            }
        }
    }

    mock_requests.post().json.return_value = response

    r = call_graphql_api(query=query, variables=variables, token=token)

    assert response['data'] == r

    snapshot.assert_match(mock_requests.post.call_args_list)

##__________________________________________________________________||
def test_error(mock_requests, snapshot):

    query = '''
      {
        viewer {
          login
          followers {
            edges {
              node {
                id
              }
            }
          }
        }
      }
    '''

    token = 'token-xxx'

    response = {
        "errors": [
            {
                "type": "MISSING_PAGINATION_BOUNDARIES",
                "path": [
                    "viewer",
                    "followers"
                ],
                "locations": [
                    {
                        "line": 9,
                        "column": 5
                    }
                ],
                "message": "You must provide a `first` or `last` value to properly paginate the `followers` connection."
            }
        ]
    }
    mock_requests.post().json.return_value = response

    with pytest.raises(Exception) as e:
        call_graphql_api(query=query, token=token)

    assert response['errors'] == e.value.args[0]

##__________________________________________________________________||
def test_bad_credentials(mock_requests):

    query = '{ viewer { login name avatarUrl } }'
    token = 'token-xxx'

    response = {
        'message': 'Bad credentials',
        'documentation_url': 'https://docs.github.com/graphql'
    }
    mock_requests.post().json.return_value = response

    with pytest.raises(Exception) as e:
        call_graphql_api(query=query, token=token)

    assert response == e.value.args[0]

##__________________________________________________________________||



import pytest

import datetime

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_users):

    y = app_users

    with y.app_context():

        ops.create_field(
            field_id=1,
            name="contact",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=2,
            name="produced_by",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=3,
            name="date_produced",
            type_=ops.FieldType.Date,
        )
        ops.create_field(
            field_id=4,
            name="field_four",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=5,
            name="field_five",
            type_=ops.FieldType.UnicodeText,
        )

        ops.commit()

    with y.app_context():

        ops.create_product_type(
            type_id=1,
            name="map",
            order=2,
            indef_article="a",
            singular="map",
            plural="maps",
            icon="mdi-map",
            field_ids=[1, 2, 3],
        )
        ops.create_product_type(
            type_id=2,
            name="beam",
            order=1,
            indef_article="a",
            singular="beam",
            plural="beams",
            icon="mdi-spotlight-beam",
            field_ids=[1, 2, 3],
        )
        ops.commit()

    with y.app_context():
        ops.create_product(
            type_id=1,
            name="map1",
            attributes={3: datetime.date(2020, 2, 1)},
        )
        ops.create_product(
            type_id=1,
            name="map2",
            attributes={3: datetime.date(2020, 2, 10)},
        )
        ops.create_product(
            type_id=1,
            name="map3",
            attributes={3: datetime.date(2020, 3, 3)},
        )
        ops.commit()

    yield y


##__________________________________________________________________||



import graphene
from graphene_sqlalchemy import SQLAlchemyObjectType

from ...models import Log as LogModel

from ..connection import CountedConnection


##__________________________________________________________________||
class Log(SQLAlchemyObjectType):
    """Record of errors, etc."""

    class Meta:
        model = LogModel
        interfaces = (graphene.relay.Node,)
        connection_class = CountedConnection


##__________________________________________________________________||



from ..fragments import FRAGMENT_FIELD

MUTATION_CREATE_FIELD = '''
mutation CreateField($input: CreateFieldInput!) {
  createField(input: $input) {
    ok
    field {
      ...fragmentField
    }
  }
}
''' + FRAGMENT_FIELD



import pytest

from sqlalchemy import exc

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product, GitHubUser


##__________________________________________________________________||
def test_product(app_empty):
    app = app_empty

    type_map = ProductType(name="map")
    map1 = Product(name="map1", type_=type_map)

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one_or_none()
        assert map1 is not None
        type_map = map1.type_
        assert "map" == type_map.name
        assert [map1] == type_map.products


##__________________________________________________________________||
def test_constraint_type_required_add(app_empty):
    app = app_empty

    map1 = Product(name="map1")

    with app.app_context():
        sa.session.add(map1)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one_or_none()
        assert map1 is None


##__________________________________________________________________||
def test_constraint_type_required_delete(app_empty):
    app = app_empty

    type_map = ProductType(name="map")
    map1 = Product(name="map1", type_=type_map)

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

    # fail to delete a type with a product
    with app.app_context():
        type_map = ProductType.query.filter_by(name="map").one()
        sa.session.delete(type_map)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()

    # assert the type and the product are still there
    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one()
        assert map1 is not None
        type_map = map1.type_
        assert "map" == type_map.name
        assert [map1] == type_map.products

    # delete the product
    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one()
        sa.session.delete(map1)
        sa.session.commit()

    # assert the type still exists
    with app.app_context():
        type_map = ProductType.query.filter_by(name="map").one()
        assert type_map is not None
        assert [] == type_map.products

    # delete the type
    with app.app_context():
        type_map = ProductType.query.filter_by(name="map").one()
        sa.session.delete(type_map)
        sa.session.commit()


##__________________________________________________________________||
def test_git_hub_user(app_empty):
    app = app_empty

    user1 = GitHubUser(login="user1", git_hub_id="04:User1")
    user2 = GitHubUser(login="user2", git_hub_id="04:User2")

    type_map = ProductType(name="map")
    map1 = Product(name="map1", type_=type_map, posting_git_hub_user=user1)

    with app.app_context():
        sa.session.add(user1)
        sa.session.add(user2)
        sa.session.add(map1)
        sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one()
        user1 = GitHubUser.query.filter_by(login="user1").one()
        assert user1 == map1.posting_git_hub_user
        assert [map1] == user1.posted_products

    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one()
        user2 = GitHubUser.query.filter_by(login="user2").one()
        map1.updating_git_hub_user = user2
        sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(name="map1").one()
        user2 = GitHubUser.query.filter_by(login="user2").one()
        assert user2 == map1.updating_git_hub_user
        assert [map1] == user2.updated_products


##__________________________________________________________________||



import pytest

from async_asgi_testclient import TestClient
from a2wsgi import WSGIMiddleware

import acondbs

QUERY = "{ version }"


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_schema(app):

    app = WSGIMiddleware(app)  # convert a wsgi app to an asgi app

    headers = {
        "Content-Type:": "application/json",
        "Authorization": "Bearer token123",  # octocat
    }

    data = {"query": QUERY}
    expected = {"data": {"version": acondbs.__version__}}

    async with TestClient(app) as client:
        resp = await client.post("/graphql", json=data, headers=headers)

    assert resp.status_code == 200
    assert resp.json() == expected


##__________________________________________________________________||



import pytest
import textwrap


from ....funcs import assert_mutation

HEADERS = {"Authorization": "Bearer token123"}  # octocat


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": textwrap.dedent(
                """
                  mutation m {
                    deleteProductFilePath(pathId: 1) { ok }
                  }
                """,
            ),
        },
        {
            "query": textwrap.dedent(
                """
                  {
                    product(productId: 1001 ) {
                      name timePosted note
                      paths { edges { node { path note product { productId } } } }
                    }
                  }
                """,
            )
        },
        id="deleteProductFilePath",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_success(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):

    success = True
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": textwrap.dedent(
                """
                  mutation m {
                    deleteProductFilePath(pathId: 15) { ok }
                  }
                """,
            )
        },
        {
            "query": textwrap.dedent(
                """
                  {
                    allProductFilePaths {
                      edges {
                        node {
                          productId
                        }
                      }
                    }
                  }
                """,
            )
        },
        id="deleteProductFilePath-error",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_error(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):
    success = False
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[createProductFilePath-noSuchField] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'productId': 1001
                    }
                },
                {
                    'node': {
                        'productId': 1012
                    }
                },
                {
                    'node': {
                        'productId': 1012
                    }
                },
                {
                    'node': {
                        'productId': 1013
                    }
                },
                {
                    'node': {
                        'productId': 1070
                    }
                },
                {
                    'node': {
                        'productId': 1120
                    }
                },
                {
                    'node': {
                        'productId': 1130
                    }
                },
                {
                    'node': {
                        'productId': 1150
                    }
                },
                {
                    'node': {
                        'productId': 1002
                    }
                },
                {
                    'node': {
                        'productId': 1002
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[createProductFilePath] 1'] = {
    'data': {
        'createProductFilePath': {
            'productFilePath': {
                'path': 'nersc:/go/to/my/new_product_v1'
            }
        }
    }
}

snapshots['test_schema_success[createProductFilePath] 2'] = {
    'data': {
        'product': {
            'name': '20180101',
            'note': '- test entry',
            'paths': {
                'edges': [
                    {
                        'node': {
                            'note': '- Note 1',
                            'path': 'nersc:/go/to/my/new_product_v1',
                            'product': {
                                'productId': '1010'
                            }
                        }
                    }
                ]
            },
            'timePosted': '2018-01-01T15:32:10'
        }
    }
}



import graphene

from ...models import (
    Product as ProductModel,
    ProductType as ProductTypeModel,
    ProductRelation as ProductRelationModel,
    ProductRelationType as ProductRelationTypeModel,
    Field as FieldModel,
)

from ..filter_ import PFilterableConnectionField
from . import type_

##__________________________________________________________________||
all_products_field = PFilterableConnectionField(type_.Product.connection)
all_product_types_field = PFilterableConnectionField(type_.ProductType.connection)  # fmt: skip
all_product_relations_field = PFilterableConnectionField(type_.ProductRelation.connection)  # fmt: skip
all_product_relation_types_field = PFilterableConnectionField(type_.ProductRelationType.connection)  # fmt: skip
all_product_file_paths_field = PFilterableConnectionField(type_.ProductFilePath.connection)  # fmt: skip
all_fields_field = PFilterableConnectionField(type_.Field.connection)


##__________________________________________________________________||
def resolve_product(parent, info, **kwargs):

    filter = [getattr(ProductModel, k) == v for k, v in kwargs.items()]
    # e.g., [ProductModel.type_id == 1, ProductModel.name == 'map_001']

    return type_.Product.get_query(info).filter(*filter).one_or_none()


product_field = graphene.Field(
    type_.Product,
    product_id=graphene.Int(),
    type_id=graphene.Int(),
    name=graphene.String(),
    resolver=resolve_product,
)


##__________________________________________________________________||
def resolve_product_type(parent, info, **kwargs):
    filter = [getattr(ProductTypeModel, k) == v for k, v in kwargs.items()]
    # e.g., [ProductTypeModel.type_id == 1, ProductTypeModel.name == 'map']

    return type_.ProductType.get_query(info).filter(*filter).one_or_none()


product_type_field = graphene.Field(
    type_.ProductType,
    type_id=graphene.Int(),
    name=graphene.String(),
    resolver=resolve_product_type,
)


##__________________________________________________________________||
def resolve_product_relation(parent, info, **kwargs):
    filter = [getattr(ProductRelationModel, k) == v for k, v in kwargs.items()]
    return type_.ProductRelation.get_query(info).filter(*filter).one_or_none()


product_relation_field = graphene.Field(
    type_.ProductRelation,
    relation_id=graphene.Int(),
    resolver=resolve_product_relation,
)


##__________________________________________________________________||
def resolve_product_relation_type(parent, info, **kwargs):
    filter = [
        getattr(ProductRelationTypeModel, k) == v for k, v in kwargs.items()
    ]
    return (
        type_.ProductRelationType.get_query(info).filter(*filter).one_or_none()
    )


product_relation_type_field = graphene.Field(
    type_.ProductRelationType,
    type_id=graphene.Int(),
    name=graphene.String(),
    resolver=resolve_product_relation_type,
)


##__________________________________________________________________||
def resolve_field(parent, info, **kwargs):

    filter = [getattr(FieldModel, k) == v for k, v in kwargs.items()]
    # e.g., [FieldModel.field_id == 1, FieldModel.name == 'contact']

    return type_.Field.get_query(info).filter(*filter).one_or_none()


field_field = graphene.Field(
    type_.Field,
    field_id=graphene.Int(),
    name=graphene.String(),
    resolver=resolve_field,
)


##__________________________________________________________________||



"""SQLAlchemy and DB related

This package contains functions, classes, and other objects that are
related to SQLAlchemy and the DB except ORM model declarations.

"""


from pathlib import Path
from flask_migrate import Migrate

from .sa import sa
from .conn import close_db_connection
from .cmds import init_db_command
from .cmds import dump_db_command
from .cmds import import_csv_command
from .cmds import export_csv_command
from .cmds import backup_db_command

migrate = Migrate()

_MIGRATIONS_DIR = Path(__file__).resolve().parent.parent.joinpath("migrations")


##__________________________________________________________________||
def init_app(app):
    """Initialize the Flask application object

    This function is called by `create_app()` of Flask

    Parameters
    ----------
    app : Flask
        The Flask application object, an instance of `Flask`
    """
    sa.init_app(app)
    migrate.init_app(app, sa, directory=_MIGRATIONS_DIR)
    app.cli.add_command(init_db_command)
    app.cli.add_command(dump_db_command)
    app.cli.add_command(import_csv_command)
    app.cli.add_command(export_csv_command)
    app.cli.add_command(backup_db_command)
    app.teardown_appcontext(close_db_connection)


##__________________________________________________________________||



"""change enum

Revision ID: 097c6676314c
Revises: 12cc2bed191c
Create Date: 2021-09-22 15:43:27.796307

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "097c6676314c"
down_revision = "12cc2bed191c"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("field", schema=None) as batch_op:
        batch_op.alter_column(
            "type_",
            type_=sa.Enum(
                "UnicodeText",
                "Boolean",
                "Integer",
                "Float",
                "Date",
                "DateTime",
                "Time",
                name="fieldtype",
            ),
            existing_type=sa.Enum(
                "UnicodeText",
                "Boolean",
                "Int",
                "Float",
                "Date",
                "DateTime",
                "Time",
                name="fieldtype",
            ),
        )


def downgrade():
    pass



import pytest

from ....funcs import assert_mutation

from ...gql import (
    FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION,
    MUTATION_UPDATE_PRODUCT_RELATION_TYPE,
)

##__________________________________________________________________||
QEURY = (
    """
{
  allProductRelationTypes {
   ...fragmentProductRelationTypeConnection
  }
}
"""
    + FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION
)

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": MUTATION_UPDATE_PRODUCT_RELATION_TYPE,
            "variables": {
                "typeId": 1,
                "input": {
                    "indefArticle": "an",
                    "singular": "mmap",
                    "plural": "mmaps",
                },
            },
        },
        {"query": QEURY},
        id="update",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_success(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):

    success = True
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": MUTATION_UPDATE_PRODUCT_RELATION_TYPE,
            "variables": {
                "typeId": 1,
                "input": {
                    "name": "mmap",
                },
            },
        },
        {"query": QEURY},
        id="error-immutable-field",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_error(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):
    success = False
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



# fmt: off
from .query_all_fields import QUERY_ALL_FIELDS  # noqa: F401
from .query_all_products import QUERY_ALL_PRODUCTS  # noqa: F401
from .query_all_products_shallow import QUERY_ALL_PRODUCTS_SHALLOW  # noqa: F401
from .query_all_products_total_count import QUERY_ALL_PRODUCTS_TOTAL_COUNT  # noqa: F401
from .query_all_product_file_paths import QUERY_ALL_PRODUCT_FILE_PATHS  # noqa: F401
from .query_all_product_relations import QUERY_ALL_PRODUCT_RELATIONS  # noqa: F401
from .query_all_product_relations_total_count import QUERY_ALL_PRODUCT_RELATIONS_TOTAL_COUNT  # noqa: F401
from .query_all_product_relation_types import QUERY_ALL_PRODUCT_RELATION_TYPES  # noqa: F401
from .query_all_product_relation_types_total_count import QUERY_ALL_PRODUCT_RELATION_TYPES_TOTAL_COUNT  # noqa: F401
from .query_all_product_types import  QUERY_ALL_PRODUCT_TYPES  # noqa: F401
from .query_field import QUERY_FIELD  # noqa: F401
from .query_product import QUERY_PRODUCT  # noqa: F401
from .query_product_shallow import QUERY_PRODUCT_SHALLOW  # noqa: F401
from .query_product_relation import QUERY_PRODUCT_RELATION  # noqa: F401
from .query_product_relation_type import QUERY_PRODUCT_RELATION_TYPE  # noqa: F401
from .query_product_type import QUERY_PRODUCT_TYPE  # noqa: F401



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import GitHubToken, GitHubUser, AccountAdmin


##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri = "sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y


@pytest.fixture
def app_users(app_empty):
    y = app_empty

    # user1: octocat - admin
    user1 = GitHubUser(login="octocat", git_hub_id="04:User583231")
    token1 = GitHubToken(  # noqa: F841
        token="90b2ee5fed25506df04fd37343bb68d1803dd97f", scope="", user=user1
    )
    admin1 = AccountAdmin(git_hub_login="octocat")

    # user2: dojocat - not admin
    user2 = GitHubUser(login="dojocat", git_hub_id="04:User9758946")
    token2 = GitHubToken(  # noqa: F841
        token="0fb8c9e16d6f7c4961c4c49212bf197d79f14080", scope="", user=user2
    )

    with y.app_context():
        sa.session.add(user1)
        sa.session.add(admin1)
        sa.session.add(user2)
        sa.session.commit()
    yield y


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[deep] 1'] = {
    'data': {
        'product': {
            'attributesBoolean': {
                'edges': [
                ]
            },
            'attributesDate': {
                'edges': [
                    {
                        'node': {
                            'field': {
                                'name': 'date_produced',
                                'type_': 'DATE'
                            },
                            'typeFieldAssociation': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                }
                            },
                            'value': '2020-02-01'
                        }
                    }
                ]
            },
            'attributesDateTime': {
                'edges': [
                ]
            },
            'attributesFloat': {
                'edges': [
                ]
            },
            'attributesInteger': {
                'edges': [
                ]
            },
            'attributesTime': {
                'edges': [
                ]
            },
            'attributesUnicodeText': {
                'edges': [
                    {
                        'node': {
                            'field': {
                                'name': 'contact',
                                'type_': 'UNICODE_TEXT'
                            },
                            'typeFieldAssociation': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                }
                            },
                            'value': None
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'produced_by',
                                'type_': 'UNICODE_TEXT'
                            },
                            'typeFieldAssociation': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                }
                            },
                            'value': None
                        }
                    }
                ]
            },
            'name': 'map1',
            'note': None,
            'paths': {
                'edges': [
                    {
                        'node': {
                            'note': None,
                            'path': 'site2:/another/way/map1',
                            'pathId': '1'
                        }
                    },
                    {
                        'node': {
                            'note': 'sample comment',
                            'path': 'site1:/path/to/map1',
                            'pathId': '8'
                        }
                    }
                ]
            },
            'postingGitHubUser': None,
            'productId': '1',
            'relations': {
                'edges': [
                    {
                        'node': {
                            'other': {
                                'name': 'beam1',
                                'productId': '4',
                                'typeId': 2,
                                'type_': {
                                    'name': 'beam',
                                    'typeId': '2'
                                }
                            },
                            'otherProductId': 4,
                            'relationId': '1',
                            'reverse': {
                                'relationId': '2',
                                'typeId': 1,
                                'type_': {
                                    'name': 'parent',
                                    'typeId': '1'
                                }
                            },
                            'reverseRelationId': 2,
                            'typeId': 2,
                            'type_': {
                                'name': 'child',
                                'typeId': '2'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'beam2',
                                'productId': '5',
                                'typeId': 2,
                                'type_': {
                                    'name': 'beam',
                                    'typeId': '2'
                                }
                            },
                            'otherProductId': 5,
                            'relationId': '3',
                            'reverse': {
                                'relationId': '4',
                                'typeId': 1,
                                'type_': {
                                    'name': 'parent',
                                    'typeId': '1'
                                }
                            },
                            'reverseRelationId': 4,
                            'typeId': 2,
                            'type_': {
                                'name': 'child',
                                'typeId': '2'
                            }
                        }
                    }
                ]
            },
            'timePosted': '2021-01-04T14:32:20',
            'timeUpdated': None,
            'typeId': 1,
            'type_': {
                'name': 'map',
                'typeId': '1'
            },
            'updatingGitHubUser': None
        }
    }
}

snapshots['test_schema[product_id-name-nonexistent] 1'] = {
    'data': {
        'product': None
    }
}

snapshots['test_schema[product_id-name] 1'] = {
    'data': {
        'product': {
            'name': 'map1',
            'productId': '1',
            'typeId': 1
        }
    }
}

snapshots['test_schema[product_id-nonexistent] 1'] = {
    'data': {
        'product': None
    }
}

snapshots['test_schema[type_id-name] 1'] = {
    'data': {
        'product': {
            'name': 'map1',
            'productId': '1',
            'typeId': 1
        }
    }
}



import pytest

from ...funcs import assert_query

from ..gql import (
    QUERY_ALL_PRODUCTS,
    QUERY_ALL_PRODUCTS_SHALLOW,
    QUERY_ALL_PRODUCTS_TOTAL_COUNT,
)

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


##__________________________________________________________________||
params = [
    pytest.param(
        {
            #
            "query": QUERY_ALL_PRODUCTS
        },
        id="deep",
    ),
    pytest.param(
        {
            #
            "query": QUERY_ALL_PRODUCTS_SHALLOW,
            "variables": {"first": 2},
        },
        id="first-two",
    ),
    pytest.param(
        {
            "query": QUERY_ALL_PRODUCTS_SHALLOW,
            "variables": {
                #
                "first": 2,
                "sort": "TIME_POSTED_DESC",
            },
        },
        id="first-two-sort",
    ),
    pytest.param(
        {
            #
            "query": QUERY_ALL_PRODUCTS_SHALLOW,
            "variables": {
                #
                "filters": {"typeId": 1},
                "first": 2,
            },
        },
        id="filters-type_id-first-two",
    ),
    pytest.param(
        {
            #
            "query": QUERY_ALL_PRODUCTS_SHALLOW,
            "variables": {
                #
                "filters": {"typeName": "map"},
                "first": 2,
            },
        },
        id="filters-type_name-first-two",
    ),
    pytest.param(
        {
            "query": QUERY_ALL_PRODUCTS_SHALLOW,
            "variables": {
                #
                "filters": {"typeId": 1},
                "sort": "TIME_POSTED_DESC",
            },
        },
        id="filters-type_id-sort",
    ),
    pytest.param(
        {
            #
            "query": QUERY_ALL_PRODUCTS_SHALLOW,
            "variables": {
                #
                "filters": {"typeName": "map"},
                "sort": "TIME_POSTED_DESC",
            },
        },
        id="filters-type_name-sort",
    ),
    pytest.param(
        {
            #
            "query": QUERY_ALL_PRODUCTS_TOTAL_COUNT,
            "variables": {
                #
                "filters": {"typeName": "map"},
                "first": 2,
            },
        },
        id="total-count",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)


##__________________________________________________________________||



import pytest
import unittest.mock as mock

from acondbs.github.ops import exchange_code_for_token

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_call(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.github.ops.call", y)
    yield y

##__________________________________________________________________||
def test_call(app, snapshot, mock_call):
    code = "kp5b8653"
    return_value = {'access_token': "wuk5phc8", 'token_type': 'bearer', 'scope': 'user'}
    mock_call.exchange_code_for_token.return_value = dict(return_value)
    with app.app_context():
        ret = exchange_code_for_token(code)
    assert ret == return_value
    snapshot.assert_match(mock_call.exchange_code_for_token.call_args_list)

##__________________________________________________________________||



from ..fragments import FRAGMENT_PRODUCT_TYPE

QUERY_PRODUCT_TYPE = (
    """
query ProductType($typeId: Int, $name: String) {
  productType(typeId: $typeId, name: $name) {
    ...fragmentProductType
  }
}
"""
    + FRAGMENT_PRODUCT_TYPE
)



"""add three columns to product_relation_types

Revision ID: 0f7fd06aa729
Revises: 0cce33011766
Create Date: 2020-05-25 20:12:35.109293

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0f7fd06aa729'
down_revision = '0cce33011766'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relation_types', schema=None) as batch_op:
        batch_op.add_column(sa.Column('indef_article', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('plural', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('singular', sa.Text(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relation_types', schema=None) as batch_op:
        batch_op.drop_column('singular')
        batch_op.drop_column('plural')
        batch_op.drop_column('indef_article')

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[one] 1'] = {
    'data': {
        'allGitHubOrgs': {
            'edges': [
                {
                    'node': {
                        'login': 'org1',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'member': {
                                            'login': 'user1'
                                        }
                                    }
                                }
                            ],
                            'totalCount': 1
                        }
                    }
                },
                {
                    'node': {
                        'login': 'org2',
                        'memberships': {
                            'edges': [
                                {
                                    'node': {
                                        'member': {
                                            'login': 'user1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'member': {
                                            'login': 'user2'
                                        }
                                    }
                                }
                            ],
                            'totalCount': 2
                        }
                    }
                },
                {
                    'node': {
                        'login': 'org3',
                        'memberships': {
                            'edges': [
                            ],
                            'totalCount': 0
                        }
                    }
                }
            ],
            'totalCount': 3
        }
    }
}



import pytest

from ...funcs import assert_query

QUERY = "{ gitHubViewer { login name avatarUrl } }"

##__________________________________________________________________||
params = [
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer token1"},
        id="one",
    ),
]


@pytest.mark.parametrize("data, headers", params)
@pytest.mark.asyncio
async def test_success(app, snapshot, data, headers):
    await assert_query(app, snapshot, data, headers)


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer no-such-token"},
        id="wrong-token",
    ),
    pytest.param(
        {"query": QUERY},
        {},
        id="no-token",
    ),
]


@pytest.mark.parametrize("data, headers", params)
@pytest.mark.asyncio
async def test_error(app, snapshot, data, headers):
    await assert_query(app, snapshot, data, headers, error=True)


##__________________________________________________________________||



from .product import (  # noqa: F401
    CreateProduct,
    UpdateProduct,
    DeleteProduct,
    ConvertProductType,
)
from .product_file_path import (  # noqa: F401
    CreateProductFilePath,
    UpdateProductFilePath,
    DeleteProductFilePath,
)
from .product_type import (  # noqa: F401
    CreateProductType,
    UpdateProductType,
    DeleteProductType,
)
from .product_relation_type import (  # noqa: F401
    CreateProductRelationTypes,
    UpdateProductRelationType,
    DeleteProductRelationTypes,
)
from .product_relation import (  # noqa: F401
    CreateProductRelation,
    DeleteProductRelation,
)
from .field import (  # noqa: F401
    CreateField,
    UpdateField,
    DeleteField,
)



"""make nullable false

Revision ID: 8e3f13ce98cb
Revises: cceebfa64b83
Create Date: 2021-11-03 19:43:59.841522

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8e3f13ce98cb'
down_revision = 'cceebfa64b83'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('type_field_association', schema=None) as batch_op:
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('type_field_association', schema=None) as batch_op:
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.alter_column('field_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('type_field_association_iid',
               existing_type=sa.INTEGER(),
               nullable=True)

    # ### end Alembic commands ###



from flask import json

import pytest
import unittest.mock as mock

from acondbs import create_app
from acondbs.db.ops import define_tables


##__________________________________________________________________||
QUERY = '''
{
  __schema {
    queryType {
      fields {
        name
      }
    }
    mutationType {
      fields {
        name
      }
    }
    subscriptionType {
      fields {
        name
      }
    }
  }
}
'''


##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri = "sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y


##__________________________________________________________________||
@pytest.fixture()
def mock_auth(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.blueprint.auth", y)
    yield y


##__________________________________________________________________||
@pytest.mark.parametrize('is_signed_in', [True, False])
@pytest.mark.parametrize('is_admin', [True, False])
def test_schema_selection(app_empty, is_signed_in, is_admin, mock_auth, snapshot):
    app = app_empty

    mock_auth.is_signed_in.return_value = is_signed_in
    mock_auth.is_admin.return_value = is_admin

    client = app.test_client()

    response = client.get('/graphql', query_string=dict(query=QUERY))

    assert 200 == response.status_code

    result = json.loads(response.data)['data']

    un_jsonified = json.loads(response.data)
    snapshot.assert_match(un_jsonified)
    # print(json.dumps(un_jsonified, indent=2))

##__________________________________________________________________||



"""make fields in product_relations non-nullable

Revision ID: aeec123e508c
Revises: ce157c2fa22e
Create Date: 2020-05-27 21:23:58.519384

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'aeec123e508c'
down_revision = 'ce157c2fa22e'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relations', schema=None) as batch_op:
        batch_op.alter_column('other_product_id',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('self_product_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relations', schema=None) as batch_op:
        batch_op.alter_column('self_product_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('other_product_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    # ### end Alembic commands ###



import datetime
import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables
from acondbs.models import FieldType
from acondbs import ops

from acondbs.db.sa import sa
from acondbs.models import GitHubUser


##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri = "sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y


@pytest.fixture
def app(app_empty):
    y = app_empty

    with y.app_context():
        user1 = GitHubUser(user_id=1, login="user1", git_hub_id="04:User1")
        user2 = GitHubUser(user_id=2, login="user2", git_hub_id="04:User2")
        sa.session.add(user1)
        sa.session.add(user2)
        sa.session.commit()

    with y.app_context():
        ops.create_field(field_id=1, name="text_one", type_=FieldType.UnicodeText)  # fmt: skip
        ops.create_field(field_id=2, name="text_two", type_=FieldType.UnicodeText)  # fmt: skip
        ops.create_field(field_id=3, name="check1", type_=FieldType.Boolean)
        ops.create_field(field_id=4, name="check2", type_=FieldType.Boolean)
        ops.create_field(field_id=5, name="number", type_=FieldType.Integer)
        ops.create_field(field_id=6, name="value", type_=FieldType.Float)
        ops.create_field(field_id=7, name="date", type_=FieldType.Date)
        ops.create_field(field_id=8, name="date_time", type_=FieldType.DateTime)  # fmt: skip
        ops.create_field(field_id=9, name="time", type_=FieldType.Time)

        ops.commit()

    with y.app_context():
        ops.create_product_type(
            type_id=1,
            name="map",
            field_ids=[1, 3, 4, 5, 6, 8],
        )
        ops.create_product_type(
            type_id=2,
            name="beam",
            field_ids=[1, 2, 3, 7, 9],
        )
        ops.commit()

    with y.app_context():
        ops.create_product_relation_type(
            type_={"type_id": 1, "name": "parent"},
            reverse={"type_id": 2, "name": "child"},
        )
        ops.commit()

    with y.app_context():
        ops.create_product(product_id=2, type_id=2, name="beam1")
        ops.create_product(product_id=3, type_id=2, name="beam2")
        ops.create_product(product_id=4, type_id=2, name="beam3")
        ops.create_product(product_id=5, type_id=2, name="beam4")
        ops.create_product(product_id=6, type_id=2, name="beam5")
        ops.commit()

        ops.create_product(
            product_id=1,
            type_id=1,
            name="map1",
            paths=["/d/e", "/a/b/c"],
            relations=[
                {"type_id": 2, "product_id": 3},
                {"type_id": 1, "product_id": 5},
            ],
            attributes={
                1: "text attribute",
                3: False,
                5: 1056,
                8: datetime.datetime(2021, 10, 17, 11, 1, 30),
            },
        )
        ops.commit()

    yield y


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[one] 1'] = {
    'data': {
        'deleteLog': {
            'ok': True
        }
    }
}

snapshots['test_schema_success[one] 2'] = {
    'data': {
        'allLogs': {
            'edges': [
                {
                    'node': {
                        'id_': '1',
                        'level': 'DEBUG',
                        'message': 'A debug message!',
                        'time': '2021-01-04T14:32:20'
                    }
                }
            ],
            'totalCount': 1
        }
    }
}



"""Configure logging

"""
import logging
from logging.config import dictConfig
from pathlib import Path

##__________________________________________________________________||
_module_path = Path(__file__).resolve().parent.parent
# the path to the dir in which the module is installed,
# i.e., the one dir above the module path.
# e.g., /home/username/venv/lib/python3.8/site-packages

# https://docs.python.org/3/library/logging.html#logrecord-objects
_old_factory = logging.getLogRecordFactory()


def record_factory(*args, **kwargs):
    """replace the pathname (the full pathname of the source file) with
    the relative path of the source file in the package

    For example, replace
    "/home/username/venv/lib/python3.8/site-packages/acondbs/schema/auth.py"
    with
    "acondbs/schema/auth.py"

    """

    record = _old_factory(*args, **kwargs)
    try:
        record.pathname = (
            Path(record.pathname).resolve().relative_to(_module_path)
        )
    except Exception:
        pass
    return record


logging.setLogRecordFactory(record_factory)


##__________________________________________________________________||
def configure_logging():
    """configure logging

    This function needs to be called early in the program, i.e., in
    the beginning of create_app() before app.logger is accessed. (It
    seems fine to be called after "app" is created.)

    https://flask.palletsprojects.com/en/1.1.x/logging/#basic-configuration

    """

    logger_name = __name__.split(".")[0]
    # i.e., "acondbs" (__name__ = "acondbs._logging")

    logger_level = "DEBUG"  # 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
    dictConfig(
        {
            "version": 1,
            "formatters": {
                "default": {
                    "format": "{asctime} {levelname:>8s} {pathname}:{lineno}: {message}",
                    "style": "{",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                }
            },
            "handlers": {
                "wsgi": {
                    "class": "logging.StreamHandler",
                    "stream": "ext://flask.logging.wsgi_errors_stream",
                    "formatter": "default",
                }
            },
            "loggers": {
                logger_name: {"level": logger_level, "handlers": ["wsgi"]}
            },
        }
    )


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[one] 1'] = {
    'data': {
        'saveWebConfig': {
            'ok': True,
            'webConfig': {
                'id_': '1',
                'json': '''{
  "headTitle": "Saved Head Title",
  "toolbarTitle": "Saved Toolbar Title",
  "devtoolLoadingstate": false,
  "productCreationDialog": false,
  "productUpdateDialog": false,
  "productDeletionDialog": false
}'''
            }
        }
    }
}

snapshots['test_schema_success[one] 2'] = {
    'data': {
        'webConfig': {
            'id_': '1',
            'json': '''{
  "headTitle": "Saved Head Title",
  "toolbarTitle": "Saved Toolbar Title",
  "devtoolLoadingstate": false,
  "productCreationDialog": false,
  "productUpdateDialog": false,
  "productDeletionDialog": false
}'''
        }
    }
}



"""remove columns from products

Revision ID: 2ed9841987be
Revises: f97ed2cdfb29
Create Date: 2021-11-19 11:20:57.432203

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2ed9841987be'
down_revision = 'f97ed2cdfb29'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.drop_column('posted_by')
        batch_op.drop_column('produced_by')
        batch_op.drop_column('updated_by')
        batch_op.drop_column('contact')
        batch_op.drop_column('date_produced')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.add_column(sa.Column('date_produced', sa.DATE(), nullable=True))
        batch_op.add_column(sa.Column('contact', sa.TEXT(), nullable=True))
        batch_op.add_column(sa.Column('updated_by', sa.TEXT(), nullable=True))
        batch_op.add_column(sa.Column('produced_by', sa.TEXT(), nullable=True))
        batch_op.add_column(sa.Column('posted_by', sa.TEXT(), nullable=True))

    # ### end Alembic commands ###



from ..models import (
    Product,
    ProductRelation,
    ProductRelationType,
)

from ..db.sa import sa


def create_product_relation(type_id, self_product_id, other_product_id, **kwargs):
    type_ = ProductRelationType.query.filter_by(type_id=type_id).one()
    self_ = Product.query.filter_by(product_id=self_product_id).one()
    other = Product.query.filter_by(product_id=other_product_id).one()
    model = ProductRelation(type_=type_, self_=self_, other=other, **kwargs)
    sa.session.add(model)
    return model


def delete_product_relation(relation_id):
    model = ProductRelation.query.filter_by(relation_id=relation_id).one()
    sa.session.delete(model)
    return



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[delete] 1'] = {
    'data': {
        'allGitHubTokens': {
            'edges': [
                {
                    'node': {
                        'scope': 'read:org',
                        'tokenId': '1',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'tokenId': '2',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'tokenId': '3',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'tokenId': '4',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                }
            ],
            'totalCount': 4
        }
    }
}

snapshots['test_schema_success[delete] 1'] = {
    'data': {
        'deleteGitHubAdminAppToken': {
            'ok': True
        }
    }
}

snapshots['test_schema_success[delete] 2'] = {
    'data': {
        'allGitHubTokens': {
            'edges': [
                {
                    'node': {
                        'scope': '',
                        'tokenId': '2',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'tokenId': '3',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'tokenId': '4',
                        'tokenMasked': 'XXXXXXXXXXXXXXX'
                    }
                }
            ],
            'totalCount': 3
        }
    }
}



"""GraphQL queries as strings

"""

from .fragments import (  # noqa: F401
    FRAGMENT_FIELD,
    FRAGMENT_FIELD_CONNECTION,
    FRAGMENT_PRODUCT,
    FRAGMENT_PRODUCT_SHALLOW,
    FRAGMENT_PRODUCT_CONNECTION,
    FRAGMENT_PRODUCT_CONNECTION_SHALLOW,
    FRAGMENT_PRODUCT_RELATION_TYPE,
    FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION,
    FRAGMENT_PRODUCT_RELATION,
    FRAGMENT_PRODUCT_RELATION_CONNECTION,
    FRAGMENT_PRODUCT_TYPE,
    FRAGMENT_PRODUCT_TYPE_CONNECTION,
)

from .queries import (  # noqa: F401
    QUERY_ALL_FIELDS,
    QUERY_ALL_PRODUCTS,
    QUERY_ALL_PRODUCTS_SHALLOW,
    QUERY_ALL_PRODUCTS_TOTAL_COUNT,
    QUERY_ALL_PRODUCT_FILE_PATHS,
    QUERY_ALL_PRODUCT_RELATIONS,
    QUERY_ALL_PRODUCT_RELATIONS_TOTAL_COUNT,
    QUERY_ALL_PRODUCT_RELATION_TYPES,
    QUERY_ALL_PRODUCT_RELATION_TYPES_TOTAL_COUNT,
    QUERY_ALL_PRODUCT_TYPES,
    QUERY_FIELD,
    QUERY_PRODUCT,
    QUERY_PRODUCT_SHALLOW,
    QUERY_PRODUCT_RELATION,
    QUERY_PRODUCT_RELATION_TYPE,
    QUERY_PRODUCT_TYPE,
)

from .mutations import (  # noqa: F401
    MUTATION_CREATE_PRODUCT,
    MUTATION_DELETE_PRODUCT,
    MUTATION_UPDATE_PRODUCT,
    MUTATION_CREATE_PRODUCT_RELATION_TYPES,
    MUTATION_UPDATE_PRODUCT_RELATION_TYPE,
    MUTATION_DELETE_PRODUCT_RELATION_TYPES,
    MUTATION_CREATE_PRODUCT_RELATION,
    MUTATION_DELETE_PRODUCT_RELATION,
    MUTATION_CREATE_PRODUCT_TYPE,
    MUTATION_UPDATE_PRODUCT_TYPE,
    MUTATION_DELETE_PRODUCT_TYPE,
    MUTATION_CREATE_FIELD,
    MUTATION_UPDATE_FIELD,
    MUTATION_DELETE_FIELD,
    MUTATION_CONVERT_PRODUCT_TYPE,
)



from ..fragments import FRAGMENT_PRODUCT

MUTATION_CONVERT_PRODUCT_TYPE = (
    """
mutation($productId: Int!, $typeId: Int!) {
  convertProductType(productId: $productId, typeId: $typeId) {
    ok
    product {
      ...fragmentProduct
    }
  }
}
"""
    + FRAGMENT_PRODUCT
)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[name] 1'] = {
    'data': {
        'productType': {
            'fields': {
                'edges': [
                    {
                        'node': {
                            'field': {
                                'name': 'contact',
                                'type_': 'UNICODE_TEXT'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'produced_by',
                                'type_': 'UNICODE_TEXT'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'date_produced',
                                'type_': 'DATE'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    }
                ]
            },
            'icon': 'mdi-map',
            'indefArticle': 'a',
            'name': 'map',
            'order': 2,
            'plural': 'maps',
            'products': {
                'edges': [
                    {
                        'node': {
                            'name': 'map1'
                        }
                    },
                    {
                        'node': {
                            'name': 'map2'
                        }
                    },
                    {
                        'node': {
                            'name': 'map3'
                        }
                    }
                ]
            },
            'singular': 'map',
            'typeId': '1'
        }
    }
}

snapshots['test_schema[type_id-and-name-nonexistent] 1'] = {
    'data': {
        'productType': None
    }
}

snapshots['test_schema[type_id-and-name] 1'] = {
    'data': {
        'productType': {
            'fields': {
                'edges': [
                    {
                        'node': {
                            'field': {
                                'name': 'contact',
                                'type_': 'UNICODE_TEXT'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'produced_by',
                                'type_': 'UNICODE_TEXT'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'date_produced',
                                'type_': 'DATE'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    }
                ]
            },
            'icon': 'mdi-map',
            'indefArticle': 'a',
            'name': 'map',
            'order': 2,
            'plural': 'maps',
            'products': {
                'edges': [
                    {
                        'node': {
                            'name': 'map1'
                        }
                    },
                    {
                        'node': {
                            'name': 'map2'
                        }
                    },
                    {
                        'node': {
                            'name': 'map3'
                        }
                    }
                ]
            },
            'singular': 'map',
            'typeId': '1'
        }
    }
}

snapshots['test_schema[type_id-sort-products] 1'] = {
    'data': {
        'productType': {
            'icon': 'mdi-map',
            'indefArticle': 'a',
            'name': 'map',
            'order': 2,
            'plural': 'maps',
            'products': {
                'edges': [
                    {
                        'node': {
                            'name': 'map1'
                        }
                    },
                    {
                        'node': {
                            'name': 'map2'
                        }
                    },
                    {
                        'node': {
                            'name': 'map3'
                        }
                    }
                ]
            },
            'singular': 'map',
            'typeId': '1'
        }
    }
}

snapshots['test_schema[type_id] 1'] = {
    'data': {
        'productType': {
            'fields': {
                'edges': [
                    {
                        'node': {
                            'field': {
                                'name': 'contact',
                                'type_': 'UNICODE_TEXT'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'produced_by',
                                'type_': 'UNICODE_TEXT'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    },
                    {
                        'node': {
                            'field': {
                                'name': 'date_produced',
                                'type_': 'DATE'
                            },
                            'type_': {
                                'name': 'map'
                            }
                        }
                    }
                ]
            },
            'icon': 'mdi-map',
            'indefArticle': 'a',
            'name': 'map',
            'order': 2,
            'plural': 'maps',
            'products': {
                'edges': [
                    {
                        'node': {
                            'name': 'map1'
                        }
                    },
                    {
                        'node': {
                            'name': 'map2'
                        }
                    },
                    {
                        'node': {
                            'name': 'map3'
                        }
                    }
                ]
            },
            'singular': 'map',
            'typeId': '1'
        }
    }
}



from ..github.ops import get_user_for_token


##__________________________________________________________________||
def get_git_hub_viewer_from_info(info):

    auth = info.context.headers.get("Authorization")
    # e.g., 'Bearer "xxxx"', "Bearer 'xxxx'",  or 'Bearer xxxx'

    if not auth:
        raise Exception("Authorization is required")

    token = auth.split()[1].strip("\"'")
    # e.g., "xxxx"

    return get_user_for_token(token)


##__________________________________________________________________||



"""add fields to github_orgs and github_users

Revision ID: 4e8693037e72
Revises: 919b1bc74839
Create Date: 2021-01-04 18:38:11.320122

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4e8693037e72'
down_revision = '919b1bc74839'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_orgs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('avatar_url', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('url', sa.Text(), nullable=True))

    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('url', sa.Text(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_users', schema=None) as batch_op:
        batch_op.drop_column('url')

    with op.batch_alter_table('github_orgs', schema=None) as batch_op:
        batch_op.drop_column('url')
        batch_op.drop_column('avatar_url')

    # ### end Alembic commands ###



from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import declarative_mixin

from ...db.sa import sa


##__________________________________________________________________||
@declarative_mixin
class AttributeBase:
    iid = sa.Column(sa.Integer(), primary_key=True)

    @declared_attr
    def product_id(self):
        return sa.Column(
            sa.Integer(),
            sa.ForeignKey("products.product_id"),
            nullable=False,
        )

    @declared_attr
    def product(self):
        return sa.relationship(
            "Product",
            backref=sa.backref(
                self.backref_column,
                cascade="all, delete-orphan",
            ),
        )

    @declared_attr
    def type_field_association_iid(self):
        return sa.Column(
            sa.Integer(),
            sa.ForeignKey("type_field_association.iid"),
            nullable=False,
        )

    @declared_attr
    def type_field_association(self):
        return sa.relationship(
            "TypeFieldAssociation",
            backref=sa.backref(
                self.backref_column,  # TODO: need to change
                cascade="all, delete-orphan",
            ),
        )

    @declared_attr
    def field_id(self):
        return sa.Column(
            sa.Integer(),
            sa.ForeignKey("field.field_id"),
            nullable=False,
        )
        # TODO: remove, replace with type_field_association_iid

    @declared_attr
    def field(self):
        return sa.relationship(
            "Field",
            backref=sa.backref(
                self.backref_column,
                cascade="all, delete-orphan",
            ),
        )
        # TODO: remove, replace with type_field_association

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.field_name!r}: {self.value!r}>"
        )

    @property
    def field_name(self):
        try:
            return self.field.name
        except BaseException:
            return self.field


##__________________________________________________________________||
class AttributeUnicodeText(AttributeBase, sa.Model):
    __tablename__ = "attribute_unicode_text"
    backref_column = "attributes_unicode_text"
    value = sa.Column(sa.UnicodeText())


class AttributeBoolean(AttributeBase, sa.Model):
    __tablename__ = "attribute_boolean"
    backref_column = "attributes_boolean"
    value = sa.Column(sa.Boolean())


class AttributeInteger(AttributeBase, sa.Model):
    __tablename__ = "attribute_integer"
    backref_column = "attributes_integer"
    value = sa.Column(sa.Integer())


class AttributeFloat(AttributeBase, sa.Model):
    __tablename__ = "attribute_float"
    backref_column = "attributes_float"
    value = sa.Column(sa.Float())


class AttributeDate(AttributeBase, sa.Model):
    __tablename__ = "attribute_date"
    backref_column = "attributes_date"
    value = sa.Column(sa.Date())


class AttributeDateTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_date_time"
    backref_column = "attributes_date_time"
    value = sa.Column(sa.DateTime())


class AttributeTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_time"
    backref_column = "attributes_time"
    value = sa.Column(sa.Time())


##__________________________________________________________________||



"""rename column

Revision ID: 0ff075f16b94
Revises: f84493df4297
Create Date: 2021-11-02 17:07:48.428164

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0ff075f16b94'
down_revision = 'f84493df4297'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.alter_column('attribute_id', new_column_name='iid', existing_type=sa.Integer())

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.alter_column('iid', new_column_name='attribute_id', existing_type=sa.Integer())

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[query] 1'] = {
    'data': {
        'gitHubOAuthAppInfo': {
            'authorizeUrl': 'https://github.com/login/oauth/authorize',
            'clientId': 'client_id_0123456789',
            'redirectUri': 'http://localhost:8080/signin'
        }
    }
}



import pytest

from ...funcs import assert_query

QUERY = "{ isSignedIn }"


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer 90b2ee5fed25506df04fd37343bb68d1803dd97f"},
        id="octocat",
    ),
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"},
        id="dojocat",
    ),
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer 1a2d18f270df3abacfb85c5413b668f97794b4ce"},
        id="wrong-token",
    ),
    pytest.param(
        {"query": QUERY},
        {},
        id="no-token",
    ),
]


@pytest.mark.parametrize("data, headers", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data, headers):
    await assert_query(app, snapshot, data, headers)


##__________________________________________________________________||



"""rename the table github_accepted_orgs github_orgs

Revision ID: 936fc5b02f16
Revises: c0d8d6a8f071
Create Date: 2020-12-26 18:43:24.367061

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '936fc5b02f16'
down_revision = 'c0d8d6a8f071'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table('github_accepted_orgs', 'github_orgs')
    # op.create_table('github_orgs',
    # sa.Column('org_id', sa.Integer(), nullable=False),
    # sa.Column('login', sa.Text(), nullable=False),
    # sa.PrimaryKeyConstraint('org_id', name=op.f('pk_github_orgs'))
    # )
    # op.drop_table('github_accepted_orgs')
    # with op.batch_alter_table('github_org_memberships', schema=None) as batch_op:
    #     batch_op.drop_constraint('fk_github_org_memberships_org_id_github_accepted_orgs', type_='foreignkey')
    #     batch_op.create_foreign_key(batch_op.f('fk_github_org_memberships_org_id_github_orgs'), 'github_orgs', ['org_id'], ['org_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table('github_orgs', 'github_accepted_orgs')
    # with op.batch_alter_table('github_org_memberships', schema=None) as batch_op:
    #     batch_op.drop_constraint(batch_op.f('fk_github_org_memberships_org_id_github_orgs'), type_='foreignkey')
    #     batch_op.create_foreign_key('fk_github_org_memberships_org_id_github_accepted_orgs', 'github_accepted_orgs', ['org_id'], ['org_id'])

    # op.create_table('github_accepted_orgs',
    # sa.Column('org_id', sa.INTEGER(), nullable=False),
    # sa.Column('login', sa.TEXT(), nullable=False),
    # sa.PrimaryKeyConstraint('org_id', name='pk_github_accepted_orgs')
    # )
    # op.drop_table('github_orgs')
    # ### end Alembic commands ###



from acondbs.db.sa import sa
from acondbs.models import GitHubOrg, GitHubUser, GitHubOrgMembership


##__________________________________________________________________||
def test_relation(app):

    with app.app_context():
        membership = GitHubOrgMembership.query.one()
        org1 = GitHubOrg.query.one()
        user1 = GitHubUser.query.one()
        assert org1 == membership.org
        assert user1 == membership.member
        assert [membership] == org1.memberships
        assert [membership] == user1.memberships


##__________________________________________________________________||
def test_cascade_delete_org(app):

    with app.app_context():
        org1 = GitHubOrg.query.filter_by(login="org1").one()
        sa.session.delete(org1)
        sa.session.commit()

    with app.app_context():
        memberships = GitHubOrgMembership.query.all()
        orgs = GitHubOrg.query.all()
        users = GitHubUser.query.all()
        assert len(memberships) == 0
        assert len(orgs) == 0
        assert len(users) == 1


def test_cascade_delete_user(app):

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="user1").one()
        sa.session.delete(user1)
        sa.session.commit()

    with app.app_context():
        memberships = GitHubOrgMembership.query.all()
        orgs = GitHubOrg.query.all()
        users = GitHubUser.query.all()
        assert len(memberships) == 0
        assert len(orgs) == 1
        assert len(users) == 0


def test_cascade_delete_membership(app):

    with app.app_context():
        membership = GitHubOrgMembership.query.one()
        sa.session.delete(membership)
        sa.session.commit()

    with app.app_context():
        memberships = GitHubOrgMembership.query.all()
        orgs = GitHubOrg.query.all()
        users = GitHubUser.query.all()
        assert len(memberships) == 0
        assert len(orgs) == 1
        assert len(users) == 1


##__________________________________________________________________||
